{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "<h1>Computation of $\\mathcal{S}(G)$</h1>\n",
    "\n",
    "<em>Pierre Guillot</em>\n",
    "\n",
    "This computes the group $\\mathcal{S}(G)$, described in my paper <em>The Grothendieck-Teichmüller group of a finite group and $G$-dessins d'enfants</em>. When $G$ is nonabelian and simple, this agrees with $\\mathcal{GT}_1(G)$.\n",
    "\n",
    "In order to use and modify the script, you first need to:\n",
    "<ul><li>create an accound on SageMathCloud, if you haven't got one already, </li>\n",
    "<li>create at least one \"project\" on SageMathCloud.</li></ul>\n",
    "\n",
    "Then you can click on \"Copy Jupyter Notebook to your project...\", in the top left hand corner.\n",
    "\n",
    "Alternatively, if you have a recent version of Sage installed, with the Jupyter Notebook working, you can click on \"Download\"."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {
    "collapsed": true
   },
   "outputs": [],
   "source": [
    "IsContained= libgap.function_factory(\"\\in\")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {
    "collapsed": false
   },
   "outputs": [],
   "source": [
    "class S_Group:\n",
    "    \n",
    "    def __init__(self, G, verbose= False, very_verbose= False):\n",
    "        \n",
    "        self.G= G\n",
    "        self.GG= GG= G.DirectProduct(G)\n",
    "        self.i1= GG.Embedding(1)\n",
    "        self.i2= GG.Embedding(2)\n",
    "        self.p1= GG.Projection(1)\n",
    "        self.p2= GG.Projection(2)\n",
    "        \n",
    "        self.verbose= verbose or very_verbose\n",
    "        self.very_verbose= very_verbose\n",
    "        \n",
    "        \n",
    "    def compute_generating_pairs(self):\n",
    "        \"\"\"Computing pairs of generators up to conjugacy.\n",
    "        \n",
    "        This builds the list of all pairs (x, y) (as elements of GG) generating G, up to the diagonal conjugation action of G.\n",
    "        Saves as self.gens_classes\n",
    "        \"\"\"\n",
    "           \n",
    "        #diagonal copy of G within G x G\n",
    "        gensG= self.G.GeneratorsOfGroup()\n",
    "        diagonal_gens= [ self.i1.Image(g) * self.i2.Image(g)  for g in gensG ]\n",
    "        diagonal_emb= self.G.GroupHomomorphismByImages(self.GG, gensG, diagonal_gens);\n",
    "        diagG= diagonal_emb.Image()\n",
    "        \n",
    "        #conj classes of pairs\n",
    "        conj_classes= diagG.OrbitsDomain(self.GG)\n",
    "        \n",
    "        if self.verbose:\n",
    "            print \"***\", len(conj_classes), \"conjugacy classes of pairs\"\n",
    "        \n",
    "        #keep only the generating pairs\n",
    "        \n",
    "        def is_generating_pair(C):\n",
    "            pair= C[0]\n",
    "            x= self.p1.Image(pair)\n",
    "            y= self.p2.Image(pair)\n",
    "            \n",
    "            return self.G.Subgroup([x, y]) == self.G\n",
    "        \n",
    "        self.gens_classes= [ C for C in conj_classes if is_generating_pair(C) ]\n",
    "\n",
    "        if self.verbose:\n",
    "            print \"***\", len(self.gens_classes), \"of them generate G\"\n",
    "    \n",
    "    \n",
    "    \n",
    "    def compute_orbits_on_pairs(self):\n",
    "        \"\"\" Orbits of G x G on pairs\n",
    "        \n",
    "        This builds a list of lists. The elements of the inner lists are indices, indicating the elements \n",
    "        in gens_classes belonging to the same G x G orbit.\n",
    "        \n",
    "        Saves as self.GG_orbits\n",
    "        \"\"\"\n",
    "    \n",
    "        if self.verbose:\n",
    "            print \"*** computing orbits of GxG\"\n",
    "    \n",
    "    \n",
    "        #group them up\n",
    "        indices= range(len(self.gens_classes))\n",
    "        partition= []\n",
    "        \n",
    "        while len(indices) > 0:\n",
    "            \n",
    "            if self.very_verbose:\n",
    "                print \"computing GxG orbits,\", len(indices), \"indices remaining\"\n",
    "            \n",
    "            index= indices[0]\n",
    "            pair= self.gens_classes[index][0] # need [0] to take a representative\n",
    "            o= self.GG.ConjugacyClass(pair).AsList()\n",
    "            related= [i for i in indices if self.gens_classes[i][0] in o]\n",
    "            # clean up the list of indices\n",
    "            indices= [i for i in indices if i not in related]\n",
    "            # save\n",
    "            partition.append( related )\n",
    "            \n",
    "        self.GG_orbits= partition # careful! numbered from 0 !\n",
    "        \n",
    "\n",
    "    def compute_theta(self):\n",
    "        \"\"\" permutation associated to theta\n",
    "        \n",
    "        This computes the permutation induced by the operation (x, y) -> (y, x) on pairs in GG. \n",
    "        Saves as self.theta, a list of pairs (=cycles of length 2), numbered from 0\n",
    "        \"\"\"\n",
    "    \n",
    "    \n",
    "        if self.verbose:\n",
    "            print \"*** computing theta\"\n",
    "    \n",
    "        indices= range(len(self.gens_classes))\n",
    "        pairs= []\n",
    "\n",
    "        while len(indices) > 0:\n",
    "            #info\n",
    "            if self.very_verbose:\n",
    "                print \"computing theta,\", len(indices), \"remaining\"\n",
    "            \n",
    "            index= indices[0]\n",
    "            pair= self.gens_classes[index][0]\n",
    "            #now let's reverse it\n",
    "            x= self.p1.Image(pair)\n",
    "            y= self.p2.Image(pair)\n",
    "            reverse= self.i1.Image(y) * self.i2.Image(x)    #pair^(p2*i1) * pair^(p1*i2);\n",
    "\n",
    "            #now where is it in the list?\n",
    "            found= False\n",
    "            i= 0\n",
    "            while not(found):\n",
    "                if reverse in self.gens_classes[i]:\n",
    "                    found= True\n",
    "                else:\n",
    "                    i= i+1\n",
    "\n",
    "            # reverse is in position i\n",
    "            indices.remove(i)\n",
    "            # have an actual cycle?\n",
    "            if index != i:\n",
    "                pairs.append( (index, i) )\n",
    "                indices.remove(index)\n",
    "\n",
    "        self.theta= pairs\n",
    "        \n",
    "\n",
    "    def compute_delta(self):\n",
    "        \"\"\" permutation associated to delta\n",
    "        \n",
    "        This computes the permutation induced by the operation (x, y) -> (y^(-1)x^(-1), y) on pairs in GG. \n",
    "        Saves as self.delta, a list of pairs (=cycles of length 2), numbered from 0\n",
    "        \"\"\"\n",
    "    \n",
    "    \n",
    "        if self.verbose:\n",
    "            print \"*** computing delta\"\n",
    "    \n",
    "        indices= range(len(self.gens_classes))\n",
    "        pairs= []\n",
    "\n",
    "        while len(indices) > 0:\n",
    "            #info\n",
    "            if self.very_verbose:\n",
    "                print \"computing delta,\", len(indices), \"remaining\"\n",
    "            \n",
    "            index= indices[0]\n",
    "            pair= self.gens_classes[index][0]\n",
    "            #now let's reverse it\n",
    "            x= self.p1.Image(pair)\n",
    "            y= self.p2.Image(pair)\n",
    "            reverse= self.i1.Image(y^(-1)*x^(-1)) * self.i2.Image(y)    #pair^(p2*i1) * pair^(p1*i2);\n",
    "\n",
    "            #now where is it in the list?\n",
    "            found= False\n",
    "            i= 0\n",
    "            while not(found):\n",
    "                if reverse in self.gens_classes[i]:\n",
    "                    found= True\n",
    "                else:\n",
    "                    i= i+1\n",
    "\n",
    "            # reverse is in position i\n",
    "            indices.remove(i)\n",
    "            # have an actual cycle?\n",
    "            if index != i:\n",
    "                pairs.append( (index, i) )\n",
    "                indices.remove(index)\n",
    "\n",
    "        self.delta= pairs\n",
    "\n",
    "        \n",
    "        \n",
    "    def permutation_associated_to(self, f):\n",
    "        \"\"\" permutation associated to the operation f on pairs. \n",
    "        \n",
    "        For example for theta one takes f= lambda x, y : (y, x). In general f(x, y) must be a pair.\n",
    "        This returns the list of images of the indices in self.gens_classes, under the operation f.\n",
    "        \n",
    "        NOTE: as suggested, it is possible to use this to compute theta, for example,\n",
    "        with f= lambda x, y : (y, x). While this is a good consistency test, it is\n",
    "        best to call self.compute_theta (or delta), which uses the fact that theta is\n",
    "        of order two (speeding up things by a factor of 2).\n",
    "        \n",
    "        This is normally used for the elements of Out(G).\n",
    "        \"\"\"\n",
    "    \n",
    "        if self.verbose:\n",
    "            print \"*** computing a permutation\"\n",
    "    \n",
    "    \n",
    "        indices= range(len(self.gens_classes))\n",
    "        images= []\n",
    "\n",
    "        for k in indices:\n",
    "            \n",
    "            if self.very_verbose:\n",
    "                print len(indices)-k, \"indices remaining\"\n",
    "            \n",
    "            \n",
    "            pair= self.gens_classes[k][0]\n",
    "            #now let's apply f\n",
    "            x= self.p1.Image(pair)\n",
    "            y= self.p2.Image(pair)\n",
    "            newpair= f(x, y)\n",
    "            xprime= newpair[0]\n",
    "            yprime= newpair[1]\n",
    "            reverse= self.i1.Image(xprime) * self.i2.Image(yprime) #name 'reverse' thinking about theta\n",
    "\n",
    "            #now where is it in the list?\n",
    "            found= False\n",
    "            i= 0\n",
    "            while not(found):\n",
    "                if reverse in self.gens_classes[i]:\n",
    "                    found= True\n",
    "                else:\n",
    "                    i= i+1\n",
    "\n",
    "            # reverse is in position i\n",
    "            images.append(i)\n",
    "        \n",
    "        return images\n",
    "\n",
    "    def compute_gens_outG(self):\n",
    "        \"\"\" Generators for the outer automorphism group.\n",
    "\n",
    "        Saved as self.gens_outG.\n",
    "        \"\"\"\n",
    "        autG= self.G.AutomorphismGroup()\n",
    "        innG= autG.InnerAutomorphismsAutomorphismGroup()\n",
    "        gens= [ x for x in autG.GeneratorsOfGroup() if not(IsContained(x, innG)) ]\n",
    "        self.gens_outG= gens\n",
    "        \n",
    "        \n",
    "    def compute_all(self, reset= False):\n",
    "        if not(hasattr(self, \"gens_classes\")) or reset:\n",
    "            self.compute_generating_pairs()\n",
    "        if not(hasattr(self, \"GG_orbits\")) or reset:\n",
    "            self.compute_orbits_on_pairs()\n",
    "        if not(hasattr(self, \"theta\")) or reset:\n",
    "            self.compute_theta()\n",
    "        if not(hasattr(self, \"delta\")) or reset:\n",
    "            self.compute_delta()\n",
    "        if not(hasattr(self, \"gens_outG\")) or reset:\n",
    "            self.compute_gens_outG()\n",
    "        if not(hasattr(self, \"perms_outG\")) or reset:\n",
    "            self.perms_outG= []\n",
    "            for a in self.gens_outG:\n",
    "                f= lambda x, y : (a.Image(x), a.Image(y))\n",
    "                self.perms_outG.append(self.permutation_associated_to(f))\n",
    "        \n",
    "        if self.verbose:\n",
    "            print \"*** preliminaries done\"\n",
    "            \n",
    "    def setup(self):\n",
    "        \"\"\" Computes all background information, and converts everything into GAP permutations and groups.\n",
    "        \n",
    "        self.compute_all() does all the hard work, but the data computed are in raw form:\n",
    "        permutations are on [0..n] instead of [1..(n+1)], some are given as lists of pairs,\n",
    "        others as list of images of [0..n]... This does the necessary conversions.\n",
    "        \n",
    "        Once the conversions are done, memory is freed by deleting the original data. Self loses\n",
    "        some attributes, and as a result, if compute_all() or setup() are called again, then\n",
    "        automatically all computations will start from scratch.        \n",
    "        \"\"\"\n",
    "        \n",
    "        # make sure everything is ready:\n",
    "        self.compute_all() # results are cached so this causes no extra work if the computations are already done\n",
    "        \n",
    "        #convenient:\n",
    "        def add_one(L):\n",
    "            return tuple( x+1 for x in L )\n",
    "        \n",
    "        # GxG orbits\n",
    "        self.blocks= [ add_one(orb) for orb in self.GG_orbits ]\n",
    "        # theta\n",
    "        pairs= [ add_one(pair) for pair in self.theta]\n",
    "        theta= libgap.eval( Permutation(pairs).cycle_string() )\n",
    "        # delta\n",
    "        pairs= [ add_one(pair) for pair in self.delta]\n",
    "        delta= libgap.eval( Permutation(pairs).cycle_string() )\n",
    "        # outG\n",
    "        perms= [ libgap.PermList(add_one(p)) for p in self.perms_outG ]\n",
    "        # group H\n",
    "        ell= len(self.gens_classes)\n",
    "        sym= libgap.SymmetricGroup(ell)\n",
    "        self.H= sym.Subgroup([theta, delta] + perms )\n",
    "        self.ell= ell\n",
    "        \n",
    "        #free some memory\n",
    "        # only self.H and self.blocks (and self.ell) are needed for factors()\n",
    "        delattr(self, \"gens_classes\")\n",
    "        delattr(self, \"GG_orbits\")\n",
    "        delattr(self, \"theta\")\n",
    "        delattr(self, \"delta\")\n",
    "        delattr(self, \"gens_outG\")\n",
    "        delattr(self, \"perms_outG\")\n",
    "        \n",
    "        if self.verbose:\n",
    "            print \"*** setup complete\"\n",
    "        \n",
    "        \n",
    "    def factors(self, reset= False):\n",
    "        \"\"\"Returns a list of groups, the direct product of which is the S-group of self.G\"\"\"\n",
    "        \n",
    "        #check if the results were cached earlier:\n",
    "        if hasattr(self, \"factors_list\") and not(reset):\n",
    "            return self.factors_list\n",
    "\n",
    "        # check whether the 'background information' is available\n",
    "        if not(hasattr(self, \"H\")):\n",
    "            self.setup() \n",
    "        \n",
    "        if self.verbose:\n",
    "            print \"*** computing the factor groups\"\n",
    "        \n",
    "        #compute the orbits of H\n",
    "        orbs= self.H.OrbitsDomain([1..self.ell])\n",
    "\n",
    "        #we gather information on these orbits\n",
    "        info_orbs= [] # this will be a list of dictionnaries\n",
    "        conj_cls_H= self.H.ConjugacyClassesSubgroups()\n",
    "        nb_classes= len(conj_cls_H)\n",
    "\n",
    "        for orb in orbs:  # for each orbit, record a few things\n",
    "    \n",
    "            #compute the stabilizer of an element\n",
    "            #or rather the conjugacy class of this stabilizer\n",
    "            stab= self.H.Stabilizer(orb[0]) #now its class\n",
    "            for i in range(nb_classes):\n",
    "                if IsContained(stab, conj_cls_H[i]):\n",
    "                    stab_num= i\n",
    "                    break\n",
    "            #ok, stab_num is the number of the conjugacy class of the stabilizer of elements of orb\n",
    "            \n",
    "            #now the sizes of the intersections of orb with all blocks:\n",
    "            # note: intersection of a list of sage integers and a list of GAP integers...\n",
    "            size_subsets= [len( set(block).intersection(set(orb.sage())) ) for block in self.blocks]\n",
    "        \n",
    "            #now store the info before moving on to next orbit\n",
    "            info_orbs.append({\"stab\" : stab_num, \"sizes\" : size_subsets})\n",
    "            \n",
    "\n",
    "        # we put together the elements of orbs according to their information in info_orbs\n",
    "        # orbits with the same info are collected in a single \"packet\" -- here a list if indices refering to orbs\n",
    "        # the union of the orbits in one packet is stable under the action of H\n",
    "\n",
    "        indices= range(len(orbs))\n",
    "        packets= []\n",
    "        \n",
    "        while len(indices) > 0:\n",
    "            i= indices[0]\n",
    "            \n",
    "            packet=  [j for j in indices if info_orbs[j] == info_orbs[i]]\n",
    "            indices= [j for j in indices if j not in packet]\n",
    "                \n",
    "            packets.append( packet )\n",
    "            \n",
    "            \n",
    "        #OK now compute small centralizers\n",
    "        gensH= self.H.GeneratorsOfGroup()\n",
    "        progress= 0\n",
    "        factors= []\n",
    "    \n",
    "        for packet in packets:\n",
    "        \n",
    "            progress= progress + 1\n",
    "            if self.very_verbose:\n",
    "                print \"progress:\", progress, \"out of\", len(packets)\n",
    "\n",
    "    \n",
    "            #find the union of the orbits in packet\n",
    "            domain= [x for i in packet for x in orbs[i]]\n",
    "            #look at the permutation group induced by H,\n",
    "            #and its centralizer\n",
    "            newgens= [g.RestrictedPermNC(domain) for g in gensH]\n",
    "            newH= libgap.Group(newgens)\n",
    "            S=    libgap.SymmetricGroup(domain)\n",
    "            Clocal= S.Centralizer(newH)\n",
    "    \n",
    "    \n",
    "            #okay now compute young subgroup \n",
    "            gens_young= [ S.Identity() ] \n",
    "            domain= [x.sage() for x in domain] # annoying: make sure we use sage integers from now on\n",
    "            for block in self.blocks:\n",
    "                subset= list( set(block).intersection(set(domain)) )\n",
    "\n",
    "                if len(subset) > 1:\n",
    "                    gens_young.extend( libgap.SymmetricGroup(subset).GeneratorsOfGroup() )\n",
    "\n",
    "            young_sub= libgap.Group( gens_young )\n",
    "    \n",
    "            #compute the intersection\n",
    "            factor= young_sub.Intersection(Clocal)\n",
    "            if not(factor.IsTrivial()):\n",
    "                \n",
    "                factors.append(factor)\n",
    "                \n",
    "                if self.very_verbose:\n",
    "                    print \"found non-trivial factor!\"\n",
    "        \n",
    "        self.factors_list= factors # cache the results\n",
    "        return factors\n",
    "\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {
    "collapsed": true
   },
   "outputs": [],
   "source": [
    "G= libgap.PSL(3, 2)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {
    "collapsed": false
   },
   "outputs": [],
   "source": [
    "S= S_Group(G, verbose= True)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "*** 197 conjugacy classes of pairs\n",
      "*** 114 of them generate G\n",
      "*** computing orbits of GxG\n",
      "*** computing theta\n",
      "*** computing delta\n",
      "*** computing a permutation\n",
      "*** preliminaries done\n",
      "*** setup complete\n",
      "*** computing the factor groups\n"
     ]
    },
    {
     "data": {
      "text/plain": [
       "[<permutation group with 1 generators>,\n",
       " <permutation group with 1 generators>,\n",
       " Group([ (14,22)(24,27), (24,27), (14,24)(22,27) ]),\n",
       " <permutation group of size 8 with 3 generators>,\n",
       " Group([ (35,40)(37,42)(44,58) ])]"
      ]
     },
     "execution_count": 6,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "S.factors()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "metadata": {
    "collapsed": false
   },
   "outputs": [],
   "source": [
    "L= S.factors() #results were cached"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
    {
     "data": {
      "text/plain": [
       "[\"C2\", \"C2\", \"D8\", \"D8\", \"C2\"]"
      ]
     },
     "execution_count": 8,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "[gp.StructureDescription() for gp in L]"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": true
   },
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "SageMath 7.3",
   "language": "",
   "name": "sagemath"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 2
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython2",
   "version": "2.7.10"
  },
  "name": "GT-simple.ipynb"
 },
 "nbformat": 4,
 "nbformat_minor": 0
}
